/**
* GSVideo library.
* Processing library based on the gstremer-java bindings by Wayne Meissner:
* http://code.google.com/p/gstreamer-java/
*
* GSCapture class for grabbing video from cameras or other capture devices.
* v0.3.11
* Changes:
* None
* Author: Andres Colubri
*/

/*
  Copyright (c) 2008 Andres Colubri

  This source is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This code is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  A copy of the GNU General Public License is available on the World
  Wide Web at <http://www.gnu.org/copyleft/gpl.html>. You can also
  obtain it by writing to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package codeanticode.gsvideo;
import processing.core.*;

import java.io.*;
import java.nio.*;
import java.lang.reflect.*;

import org.gstreamer.*;
import org.gstreamer.elements.*;

/**
 * GSCapture class.
 */
public class GSCapture extends PImage implements PConstants 
{
    Method captureEventMethod; 
  
    boolean available;
    boolean firstFrame = true;
  
    int captureWidth;
    int captureHeight;
    IntBuffer capturePixels;

    Pipeline gpipe;
 
    GSVideo gsvideo; 
 
    public GSCapture(PApplet parent, int requestWidth, int requestHeight)
    {
        super(requestWidth, requestHeight, RGB);
        this.parent = parent;
        initOS(requestWidth, requestHeight, 25);	
    }
  
    public GSCapture(PApplet parent, int requestWidth, int requestHeight, int frameRate)
    {
        super(requestWidth, requestHeight, RGB);
        this.parent = parent;
        initOS(requestWidth, requestHeight, frameRate);
    }

    public GSCapture(PApplet parent, int requestWidth, int requestHeight, String sourceName)
    {
        this(parent, requestWidth, requestHeight, sourceName, new String[] { }, new int[] { }, 25);
    }  
  
    public GSCapture(PApplet parent, int requestWidth, int requestHeight, String sourceName, int frameRate)
    {
        this(parent, requestWidth, requestHeight, sourceName, new String[] { }, new int[] { }, frameRate);
    }
  
    public GSCapture(PApplet parent, int requestWidth, int requestHeight, String sourceName, String[] propNames, int[] propValues, int frameRate)
    {
        super(requestWidth, requestHeight, RGB);
        this.parent = parent;
        init(requestWidth, requestHeight, sourceName, propNames, propValues, frameRate);
    }
  
    void initOS(int requestWidth, int requestHeight, int frameRate)
    {
        if (gsvideo.OS() == gsvideo.LINUX) 
        {
            init(requestWidth, requestHeight, "v4l2src", new String[] { }, new int[] { }, frameRate);
        }
	else if (gsvideo.OS() == gsvideo.MACOS) 
        {
            init(requestWidth, requestHeight, "qtwrappersrc", new String[] { }, new int[] { }, frameRate);
        }
        else if (gsvideo.OS() == gsvideo.WINDOWS) 
        {
            init(requestWidth, requestHeight, "dshowvideosrc", new String[] { }, new int[] { }, frameRate);
        }
	else 
        {
            init(requestWidth, requestHeight, "videotestsrc", new String[] { "pattern" }, new int[] { 0 }, frameRate);
        }
    }

    void init(int requestWidth, int requestHeight, String sourceName, String[] propNames, int[] propValues, int frameRate)
    {
        gpipe = null;
	
        gsvideo = new GSVideo();
        gsvideo.init();

        // register methods
        parent.registerDispose(this);

        try 
        {
            captureEventMethod = parent.getClass().getMethod("captureEvent", new Class[] { GSCapture.class });
        } 
        catch (Exception e) 
        {
            // no such method, or an error.. which is fine, just ignore
        }
	  
        gpipe = new Pipeline("GSCapturePipeline");
	  
        Element videoSource = ElementFactory.make(sourceName, "Source");
	
        if (propNames.length != propValues.length) 
        {
            parent.die("Error: number of property names is different from number of values.", null);
        }
	
        for (int i = 0; i < propNames.length; i++) 
        {
            videoSource.set(propNames[i], propValues[i]);
        }
	
	// conv is needed for dshowvideosrc to work, but is incompatible with matroxcapture... FIX!
	Element conv = ElementFactory.make("ffmpegcolorspace", "ColorConverter");
	
        Element videofilter = ElementFactory.make("capsfilter", "ColorFilter");
        videofilter.setCaps(new Caps("video/x-raw-rgb, width=" + requestWidth + ", height=" + requestHeight + ", bpp=32, depth=24"));
	
        RGBDataSink videoSink = new RGBDataSink("rgb",
            new RGBDataSink.Listener() 
            {
                public void rgbFrame(int w, int h, IntBuffer buffer) 
                {
                    invokeEvent(w, h, buffer);
                }
            }	
        );
	  
        gpipe.addMany(videoSource, conv, videofilter, videoSink);
        Element.linkMany(videoSource, conv, videofilter, videoSink);  
	
        //gpipe.addMany(videoSource, videofilter, videoSink);
        //Element.linkMany(videoSource, videofilter, videoSink);  	
	
        gpipe.setState(State.PLAYING);  
    }

    void invokeEvent(int w, int h, IntBuffer buffer) 
    {
        available = true;  
        captureWidth = w;
        captureHeight = h;
        capturePixels = buffer;
	  
        // Creates a movieEvent.
        if (captureEventMethod != null) 
        {
            try 
            {
                captureEventMethod.invoke(parent, new Object[] { this });
            } 
            catch (Exception e) 
            {
                System.err.println("error, disabling captureEvent() for videotestsrc");
                e.printStackTrace();
                captureEventMethod = null;
            }
        }
    }

    public void resume()
    {
        gpipe.setState(State.PLAYING);
    }
  
    public void stop()
    {
        gpipe.setState(State.PAUSED);  
    }
  
    public void read() 
    {
        if (firstFrame) 
        {
            //super.init(captureWidth, captureHeight, RGB);
            loadPixels();
            firstFrame = false;
        }		
        capturePixels.get(pixels);
        updatePixels();
	
        available = false;
    }
  
    /**
     * Get a list of all available captures as a String array.
     * i.e. println(Capture.list()) will show you the goodies.
     */  
    static public String[] list() 
    {
        String listing[] = new String[0];
        return listing;  
    }
 
    /**
     * Set the frameRate for how quickly new frames are read
     * from the capture device.
     */
    public void frameRate(int iframeRate) 
    {
    }
 
    public boolean available() 
    {
        return available;
    }
 
    public void dispose() 
    {
    } 
}
